/* 

  RMC - Ron's MUD Client
  Copyright (C) 2000 Ron Martin <rmartin@cps.cmich.edu>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
*/

/* 
   helpers.c

   Functions that help other funtions do their jobs.
   
*/

#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <unistd.h>
#include <ctype.h>
#include <regex.h>
#include <sys/stat.h>

#include "client.h"
#include "callbacks.h"
#include "constants.h"
#include "interface.h"

extern struct conn_info * info_database;
extern GdkColor colors[8];
extern GdkColor color_normal;

/*
	register_new_conn() - gets data from a 'create_new_conn'
	window and creates a connection, storing it in the info
	database and writing it to disk.
	
	It is important there is a new line after the entries so
	that it will be written to disk properly. - RPM 4/2/00
*/
void register_new_conn(GtkWidget * widget)
{
	struct conn_info * temp_info = NULL, input_info;
	GtkWidget * entry_temp, * clist;
	char *text_data[3];

	void init_clist(GtkWidget * widget);
	void write_info_database();
	
        if(!(entry_temp = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "entry_name_host")))
        {
        	error_handle(7);
        	return;
        }
        text_data[1] = gtk_entry_get_text(GTK_ENTRY(entry_temp));
	input_info.host_name = g_malloc(strlen(text_data[1]) + 2);
	strcpy(input_info.host_name, text_data[1]);
	strcat(input_info.host_name, "\n");

        if(!(entry_temp = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "entry_port")))
        {
        	error_handle(7);
        	return;
        }
	text_data[2] = gtk_entry_get_text(GTK_ENTRY(entry_temp));
	input_info.port = atoi(text_data[2]); /* port is an int */
	
        entry_temp = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "entry_name_mud");
        text_data[0] = gtk_entry_get_text(GTK_ENTRY(entry_temp));
	input_info.mud_name = g_malloc(strlen(text_data[0]) + 2);
	strcpy(input_info.mud_name, text_data[0]);	
	strcat(input_info.mud_name, "\n");
		
	temp_info = info_database;
	
	if(temp_info)
	{
		for(; temp_info; temp_info = temp_info->next)
		{
			if(strcmp(input_info.mud_name, temp_info->mud_name)) /* find duplicate names */
				continue;
			error_handle(8);
			return;
		}
		for(temp_info = info_database; temp_info->next; temp_info = temp_info->next);
		/* ^^ finds last data in info_database linked list */

		temp_info->next = g_malloc0(sizeof(conn_info));
		temp_info = temp_info->next;
	}
	else
	{
		info_database = g_malloc0(sizeof(conn_info));
		temp_info = info_database;
	}

	temp_info->mud_name = g_strdup(input_info.mud_name);
	temp_info->host_name = g_strdup(input_info.host_name);
	temp_info->port = input_info.port;
	temp_info->font_name = g_strdup("-misc-fixed-medium-r-semicondensed-*-*-120-*-*-c-*-iso8859-8");
	temp_info->font = gdk_font_load("-misc-fixed-medium-r-semicondensed-*-*-120-*-*-c-*-iso8859-8");
	write_info_database();
	
	clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "clist");
	init_clist(clist);
	
	delete_event(widget, NULL, NULL);
	
}

/*
	inti_clist() - initializes the list in the main window.
	This function sets all the text in the list window so the
	user can select which connection to utilize.  It also
	sets data for each row to point to the connection with 
	which it is associated.
*/
void init_clist(GtkWidget * widget)
{
  char * text[3], temp[16];
  struct conn_info * temp_info;
  GtkCList * clist;
  int i = 0;
  
  clist = (GtkCList *)widget;
  
  gtk_clist_clear(clist);
  
  if(info_database)
  	for(temp_info = info_database; temp_info; temp_info = temp_info->next)
  	{
  		text[0] = g_strdup(temp_info->mud_name);
  		text[1] = g_strdup(temp_info->host_name);
  		sprintf(temp, "%d", temp_info->port);
  		text[2] = g_strdup(temp);
  		gtk_clist_append(GTK_CLIST(clist), text);
  		gtk_clist_set_row_data(GTK_CLIST(clist), i, temp_info); /* sets the row data */
  		i++; /* It took me forever to realize I didn't have this here. >:( - RPM 1/31/00 */
  		g_free(text[0]);
  		g_free(text[1]);
  		g_free(text[2]);
  	}
  gtk_clist_select_row(clist, 0, 0);
  return;
}

/*
	update_conn_text() - updates the text box associated with an open connection.
	This function receives up to 8192 bytes of data from a socket, storing
	it in a buffer.  If the number of bytes < 1 then either it is EOF or
	there was an error.  Either way, disconnect.  The received data will be sent
	to insert_text for proper insertion into the text box.  The buffer will then
	be stripped of escape sequences and passed to process_trigger().
*/
void update_conn_text(gpointer output_widget, int socket_descriptor, GdkInputCondition condition)
{
	char buf[8192];
	GtkWidget * text;
	GtkWidget * window;
	GtkWidget * entry;
	struct conn_info * temp_info;
	int num_bytes;
	
	void insert_text(GtkWidget * text, GtkWidget * entry, char * buf, float value, GdkFont * font);
	void hang_up(int tag, int socket_descriptor);
	void process_trigger(struct conn_info * temp_info, char * input, GtkWidget * widget);
	void strip_escapes(char * input);

	window = (GtkWidget *)output_widget;
	temp_info = gtk_object_get_data(GTK_OBJECT(window), "conn_info");
	
	memset(buf, 0, 8192);
	
	text = gtk_object_get_data(GTK_OBJECT(window), "text");
	entry = gtk_object_get_data(GTK_OBJECT(window), "entry");
	/* ^^ so that the entry can be set to invisible if ECHO_ON is received. */
	
	num_bytes = recv(socket_descriptor, buf, 8192, 0); /* get the data */
	
	if(num_bytes < 1)
	{
		hang_up(temp_info->descriptor, socket_descriptor);
		temp_info->state &= ~STATE_CONNECTED;
		
		return;
	}

	insert_text(text, entry,  buf, temp_info->value, temp_info->font);

	strip_escapes(buf); /* for process_trigger */
	process_trigger(temp_info, buf, window); /* compare data to the triggers */
}

/*
	insert_text() - inserts given text to a given text box.
	This funciton needs to find all the ANSI color codes and
	adjust the colors accordingly.  It also needs to find the
	ECHO_OFF or ECHO_ON commads so the command entry widget can
	be set properly.  It uses a custom structure to keep track of
	the foreground and background colors, as well as wether or not
	'bold' is turned on.
*/
void insert_text(GtkWidget * text, GtkWidget * entry, char * buf, float value, GdkFont * font)
{

	int i = 0, temp = 0;
	unsigned char buf2[8192];
	GtkAdjustment * adj;
	struct color_info temp_color;
				
	void insert_text_helper(char * buf, GtkWidget * text, struct color_info color, float value, GdkFont * font);
	
	temp_color.bold = 0;
	temp_color.fg_color = 37;
	temp_color.bg_color = 40;
	
	adj = GTK_TEXT(text)->vadj;
	memset(buf2, 0, 8192);

	/* Appends a new line so it looks nice. */
	insert_text_helper("\n", text, temp_color, value, font);

	while(buf[i] != '\0') /* while we're still in the string */
	{
		if(buf[i] == '\xFF') /* telnet sequnce starting */
		{
			if(!strncmp(ECHO_ON, &buf[i], strlen(ECHO_ON))) /* is it ECHO_ON? */
			{
				gtk_entry_set_visibility(GTK_ENTRY(entry), TRUE); /* make it visible */
				gtk_entry_set_text(GTK_ENTRY(entry), ""); /* clear it so the password isn't seen */
			}

			if(!strncmp(ECHO_OFF, &buf[i], strlen(ECHO_OFF)))
			{
				gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE); /* is it ECHO_OFF? */
				gtk_entry_set_text(GTK_ENTRY(entry), ""); /* make it invisible */
			}
			i += 3; /* skip passed the sequence */
			continue;
		}
		if(buf[i] == '\x1B') /* escape */
		{
			i++;

			if(buf[i] == '[')
			{
				i++;
				
				while(1)
				{
					temp = atoi(&buf[i]);
					if(temp == 0)
					{
						temp_color.bold = 0;
						temp_color.fg_color = 37;
						temp_color.bg_color = 40;
					}
					else if (temp == 1)
						temp_color.bold = 1;
					else if(temp > 29 && temp < 38)
						temp_color.fg_color = temp;
					else if(temp > 39 && temp < 48)
						temp_color.bg_color = temp;
					
					while(isdigit(buf[i]))
						i++;
					if(buf[i] == '\x1b')
						 continue;
					if(buf[i] == 'm')
						break;
					i++;
				}
				insert_text_helper(buf2, text, temp_color, value, font);
				memset(buf2, 0, 8192);
				i++;
				continue;
			}
			i++;
			continue;
		}
		if(buf[i] != '\xD') /* catch unhandled messages */
		{
			if(buf[i] == '\xF2')
				buf2[strlen(buf2)] = '\n';
			else
				buf2[strlen(buf2)] = buf[i];
		}
		i++;
	}
	insert_text_helper(buf2, text, temp_color, value, font);
	memset(buf2, 0, 8192);
}

/*
	insert_text_helper() - inserts text into a text widget with a specified
	color.
*/
void insert_text_helper(char * buf, GtkWidget * text, struct color_info temp_color, float value, GdkFont * font)
{
	struct color_info * color;
	GtkAdjustment * adj;
	int scroll = FALSE;
	
	adj = GTK_TEXT(text)->vadj;
	
	color = (struct color_info *)gtk_object_get_data(GTK_OBJECT(text), "curr_color");
	
	if(adj->value == adj->upper - adj->lower - adj->page_size)
		scroll = TRUE;
	
	gtk_text_freeze(GTK_TEXT(text));
	gtk_text_insert(GTK_TEXT(text), font, &colors[color->fg_color - 30 + color->bold * 8], &colors[color->bg_color - 40], buf, -1);
	gtk_text_thaw(GTK_TEXT(text));
	
	color->bold = temp_color.bold;
	color->fg_color = temp_color.fg_color;
	color->bg_color = temp_color.bg_color;
	
	if(scroll)
		gtk_adjustment_set_value(adj, adj->upper - adj->lower - adj->page_size);
}

/*
	hang_up() - hangs up a connection.
	It is not only important to hang up, but also to
	remove it from gdk's input handler so it doesn't barf
	on you.
*/
void hang_up(int tag, int socket_descriptor)
{
	close(socket_descriptor);
	gdk_input_remove(tag);
	error_handle(6);
	return;
}

/*
	init_parse_clist() - initializes a clist with parse data.
	This function handles things common with an alias clist and
	a trigger clist.  It simply goes through the parse type
	linked list and sets the clist data appropriately.
*/
void init_parse_clist(struct parse_type * temp_parse, GtkWidget * clist)
{
	int i = 0;

	gtk_clist_clear(GTK_CLIST(clist));
	if(!temp_parse)
		return;
	for(; temp_parse; temp_parse = temp_parse->next)
	{
		gtk_clist_append(GTK_CLIST(clist), &temp_parse->input_string);
		gtk_clist_set_row_data(GTK_CLIST(clist), i, temp_parse);
		i++;
	}
}

/*
	init_edit_conn() - intializes the edit window with a given connection.
	This function sets all the important data in the connection, ready for
	user interaction.
*/
void init_edit_conn(GtkWidget * window, struct conn_info * temp_conn)
{
	GtkWidget * temp_widget, * clist_alias, * clist_trigger, * clist_log;
	struct conn_info * copy_conn = NULL;
	char port[16];
		
	struct conn_info * conn_dup(struct conn_info * input_conn);
	void init_parse_clist(struct parse_type * temp_parse, GtkWidget * clist);
	void init_log_clist(struct log_info * log, GtkCList * clist);
	
	copy_conn = conn_dup(temp_conn); /* so we don't f*** up the database */
	if(!copy_conn)
		return;
	gtk_object_set_data(GTK_OBJECT(window), "conn_info", copy_conn);
	gtk_object_set_data(GTK_OBJECT(window), "real_conn_info", temp_conn); /* so we can recover */
	
	temp_widget = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(window), "entry_edit_general_mud");
	gtk_entry_set_text(GTK_ENTRY(temp_widget), copy_conn->mud_name);
	
	temp_widget = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(window), "entry_edit_general_host");
	gtk_entry_set_text(GTK_ENTRY(temp_widget), copy_conn->host_name);
	
	temp_widget = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(window), "entry_edit_general_port");
	sprintf(port, "%d\n", copy_conn->port);
	gtk_entry_set_text(GTK_ENTRY(temp_widget), port);
	
	temp_widget = NULL;
	gtk_object_set_data(GTK_OBJECT(window), "curr_alias", copy_conn->aliases);
	gtk_object_set_data(GTK_OBJECT(window), "curr_trigger", copy_conn->triggers);
	gtk_object_set_data(GTK_OBJECT(window), "curr_log", copy_conn->logs);
	
	clist_alias = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(window), "clist_alias");
	clist_trigger = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(window), "clist_trigger");
	clist_log = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(window), "log_clist");
	
	init_parse_clist(copy_conn->aliases, clist_alias);
	init_parse_clist(copy_conn->triggers, clist_trigger);
	init_log_clist(copy_conn->logs, GTK_CLIST(clist_log));
	
	gtk_clist_select_row(GTK_CLIST(clist_alias), 0, 0);
	gtk_clist_select_row(GTK_CLIST(clist_trigger), 0, 0);
	gtk_clist_select_row(GTK_CLIST(clist_log), 0, 0);
}

/*
	select_row_helper() - changes the command text widget of either
	the edit alias window or edit trigger window.
*/
void select_row_helper(GtkWidget * widget, struct parse_type * temp_parse, int type)
{
	GtkWidget * temp_entry = NULL, * temp_text = NULL;

	switch(type)
	{
		case 1:
			temp_entry = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "entry_alias");
			temp_text = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "text_alias");
 			break;
		case 2:
			temp_entry = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "entry_trigger");
			temp_text = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "text_trigger");
			break;
	}
	if(!temp_parse)
		return;

	if(temp_parse->input_string)
		gtk_entry_set_text(GTK_ENTRY(temp_entry), temp_parse->input_string);
	if(temp_parse->output_string)
	{
		gtk_text_set_point(GTK_TEXT(temp_text), 0);
		gtk_text_forward_delete(GTK_TEXT(temp_text), gtk_text_get_length(GTK_TEXT(temp_text)));
		/* ^^ clear the text box */
		gtk_text_insert(GTK_TEXT(temp_text), NULL, NULL, NULL, temp_parse->output_string, -1);
	}
}

/*
	conn_dup() - duplicate a connection.
	Duplicates a given connection.  Also duplicates all
	parse_type linked lists so that all memory addresses are
	unique as compared to the given connection information.
*/
struct conn_info * conn_dup(struct conn_info * input_conn)
{
	struct conn_info * temp_conn;
	
	struct parse_type * parse_dup(struct parse_type * input_parse);
	struct log_info * log_dup(struct log_info * input_log);
	
	temp_conn = g_malloc0(sizeof(conn_info));
	
	temp_conn->host_name = g_strdup(input_conn->host_name);
	temp_conn->mud_name = g_strdup(input_conn->mud_name);
	temp_conn->port = input_conn->port;
	temp_conn->font = input_conn->font;
	temp_conn->aliases = parse_dup(input_conn->aliases); /* duplicate aliases */
	temp_conn->triggers = parse_dup(input_conn->triggers); /* and triggers */
	temp_conn->logs = log_dup(input_conn->logs);
	temp_conn->font_name = g_strdup(input_conn->font_name);
	return temp_conn;
}

/*
	parse_dup() - duplicates a parse_type linked list
	This duplicates a linked list that holds text parse information
	(aliases and triggers).  It msut traverse the entire linked list,
	create a new list fron the given information.
*/
struct parse_type * parse_dup(struct parse_type * input_parse)
{
	struct parse_type * temp_parse = NULL, * return_parse = NULL;
	
	if(!input_parse)
		return NULL;
		
	while(input_parse) /* traverse the list */
	{
		if(!temp_parse) /* first time? */
		{
			temp_parse = g_malloc0(sizeof(parse_type));
			return_parse = temp_parse; /* temp_parse will change.  return_parse won't */
									   /* needed to return */
		}
		else /* use the list we have created already */
		{
			temp_parse->next = g_malloc0(sizeof(parse_type));
			temp_parse = temp_parse->next;
		}
		
		temp_parse->input_string = g_strdup(input_parse->input_string);
		temp_parse->output_string = g_strdup(input_parse->output_string);

		input_parse = input_parse->next;
	}
	
	return return_parse;
}

struct log_info * log_dup(struct log_info * input_log)
{
	struct log_info * temp_log = NULL, * return_log = NULL;
	
	if(!input_log)
		return NULL;
		
	while(input_log)
	{
		if(!temp_log)
		{
			temp_log = g_malloc0(sizeof(log_info));
			return_log = temp_log;
		}
		else
		{
			temp_log->next = g_malloc0(sizeof(log_info));
			temp_log = temp_log->next;
		}
		
		temp_log->name = g_strdup(input_log->name);
		temp_log->log_file = g_strdup(input_log->log_file);
		temp_log->log_to_window = input_log->log_to_window;
		temp_log->log_widget = input_log->log_widget;
		input_log = input_log->next;
	}
	
	return return_log;
}

/*
	update_clist_from_entry() - updates an alias or trigger clist when the
	entry box is changed.
	This just handles some eye candy.  I though it would be neat if
	the clist changed as you typed.  I would not recommend doing this again
	if you want to add it to another application.  It was odd to
	implement and probably slows the application.
*/
void update_clist_from_entry(GtkWidget * widget, int type)
{
	GtkWidget * entry = NULL, * clist = NULL;
	struct parse_type * temp_parse = NULL;
	int row;
	
	switch(type)
	{
		case 1:
			entry = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "entry_alias");
			clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "clist_alias");
			temp_parse = (struct parse_type *)gtk_object_get_data(GTK_OBJECT(widget), "curr_alias");
			break;
		case 2:
			entry = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "entry_trigger");
			clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "clist_trigger");
			temp_parse = (struct parse_type *)gtk_object_get_data(GTK_OBJECT(widget), "curr_trigger");
			break;
	}
	
	if(temp_parse)
	{
		row = gtk_clist_find_row_from_data(GTK_CLIST(clist), temp_parse);
		gtk_clist_set_text(GTK_CLIST(clist), row, 0, gtk_entry_get_text(GTK_ENTRY(entry)));
		free(temp_parse->input_string); /* garbage collection is your friend. */
		temp_parse->input_string = strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
	}
}

/*
	edit_text_change() - used when a trigger or alias command text
	box is edited.
	This is a very simple function.  I debated whether or not to
	have it.  It's here now, so there's no turning back.
*/
void edit_text_change(GtkWidget * text, struct parse_type * input_parse, int type)
{
	if(input_parse->output_string)
		g_free(input_parse->output_string);
	
	input_parse->output_string = g_strdup(gtk_editable_get_chars(GTK_EDITABLE(text), 0, -1));
}

/*
	write_info_database() - writes the database to a file.
	Rather than having each edit operation attempt to reparse
	the conneciton file, I decided it would be much easier to
	rewrite the whole thing out to file.  The file will always
	be ~/.rmcrc.  The home directory will be ascertained via the
	getenv() function (see the man page).  The database will be
	initially written to ~./rmcrc.bak, just in case the writing
	procedure bails, the original database will not be overwritten.
*/
void write_info_database()
{
	struct conn_info * temp_info;
	struct parse_type * temp_parse;
	struct log_info * temp_log;
	FILE * rc_file;
	char  * fname = NULL, * home, * temp;

	home = getenv("HOME");
	
	fname = g_malloc(strlen(home) + 12);
	
	strcpy(fname, home);
	strcat(fname, "/.rmcrc");
	strcat(fname, ".bak");
	
	if(!(rc_file = fopen(fname, "w")))
	{
		error_handle(2);
		return;
	}

	temp_parse = NULL;

	if(info_database) /* is there data? */
		for(temp_info = info_database; temp_info; temp_info = temp_info->next) /* traverse database */
		{
			fputs("{\n", rc_file);
			fprintf(rc_file, "%s", temp_info->mud_name);
			if(temp_info->mud_name[strlen(temp_info->mud_name) - 1] != '\n')
				fprintf(rc_file, "\n"); /* make sure there are newlines! */
			fprintf(rc_file, "%s", temp_info->host_name);
			if(temp_info->host_name[strlen(temp_info->host_name) - 1] != '\n')
				fprintf(rc_file, "\n");
			fprintf(rc_file, "%d\n", temp_info->port);
			fprintf(rc_file, "%s\n", temp_info->font_name);
			
			if(temp_info->aliases) /* are there aliases? */
				for(temp_parse = temp_info->aliases; temp_parse; temp_parse = temp_parse->next)
				{   /* ^^ traverse alias list */
					fputc('\n', rc_file);
					
					fputs("alias\n", rc_file); /* beginning of alias */
					if(temp_parse->input_string[strlen(temp_parse->input_string) - 1] != '\n')
					{
						temp = temp_parse->input_string;
						temp_parse->input_string = g_malloc0(strlen(temp_parse->input_string) + 2);
						strcpy(temp_parse->input_string, temp);
						strcat(temp_parse->input_string, "\n");
						g_free(temp);
					}
					fprintf(rc_file, "%s", temp_parse->input_string);
					if(temp_parse->output_string[strlen(temp_parse->output_string) - 1] != '\n')
					{
						temp = temp_parse->output_string;
						temp_parse->output_string = g_malloc0(strlen(temp_parse->output_string) + 2);
						strcpy(temp_parse->output_string, temp);
						strcat(temp_parse->output_string, "\n");
						g_free(temp);
					}
					fprintf(rc_file, "%s", temp_parse->output_string);
					fputs("done\n", rc_file); /* end of alias */
				}
			if(temp_info->triggers) /* are there triggers? */
				for(temp_parse = temp_info->triggers; temp_parse; temp_parse = temp_parse->next)
				{
					fputc('\n', rc_file);
					
					fputs("trigger\n", rc_file); /* beginning of trigger */
					if(temp_parse->input_string[strlen(temp_parse->input_string) - 1] != '\n')
					{
						temp = temp_parse->input_string;
						temp_parse->input_string = g_malloc0(strlen(temp_parse->input_string) + 2);
						strcpy(temp_parse->input_string, temp);
						strcat(temp_parse->input_string, "\n");
						g_free(temp);
					}
					fprintf(rc_file, "%s", temp_parse->input_string);
					if(temp_parse->output_string[strlen(temp_parse->output_string) - 1] != '\n')
					{
						temp = temp_parse->output_string;
						temp_parse->output_string = g_malloc0(strlen(temp_parse->output_string) + 2);
						strcpy(temp_parse->output_string, temp);
						strcat(temp_parse->output_string, "\n");
						g_free(temp);
					}
					fprintf(rc_file, "%s", temp_parse->output_string);
					fputs("done\n", rc_file); /* end of trigger */
				}
			if(temp_info->logs)
				for(temp_log = temp_info->logs; temp_log; temp_log = temp_log->next)
				{
					fputc('\n', rc_file);
					
					fputs("log\n", rc_file);
					fputs(temp_log->name, rc_file);
					fputc('\n', rc_file);
					if(temp_log->log_file)
						fputs(temp_log->log_file, rc_file);
					else
						fputc('*', rc_file);
					fputc('\n', rc_file);
					if(temp_log->log_to_window)
						fputs("True\n", rc_file);
					else
						fputs("False\n", rc_file);
					fputs("done\n", rc_file);
				}
				
			fputs("}\n", rc_file);
		}
	
	fclose(rc_file); /* all done */
	system("rm ~/.rmcrc"); /* the ol' switcharoo */
	system("mv ~/.rmcrc.bak ~/.rmcrc");
	free(fname);

}

/*
	conn_free() - frees memory associated with a connection
	This function frees all the strings in a connection, as
	well as all the aliases and triggers, and even the conneciton
	itself.  How handy.
*/
void conn_free(struct conn_info * temp_conn)
{
	void parse_free_chain(struct parse_type * temp_parse);
	void free_log_chain(struct log_info * log);
	void free_history(struct commands * history);
	
	if(!temp_conn)
		return;
		
	parse_free_chain(temp_conn->aliases); /* free aliases */
	parse_free_chain(temp_conn->triggers); /* free triggers */
	free_log_chain(temp_conn->logs);
	free_history(temp_conn->history);
		
	if(temp_conn->mud_name)
		g_free(temp_conn->mud_name); 
	if(temp_conn->host_name)
		g_free(temp_conn->host_name);
		
	temp_conn->mud_name = NULL;
	temp_conn->host_name = NULL;
	temp_conn->next = NULL;
	temp_conn->aliases = NULL;
	temp_conn->triggers = NULL;
	
	g_free(temp_conn);
}

void free_history(struct commands * history)
{
	struct commands * temp_history;
	
	if(!history)
		return;
		
	while(history->prev)
		history = history->prev;
	
	while(history)
	{
		temp_history = history->next;
		g_free(history->command);
		g_free(history);
		
		history = temp_history;
	}
}

/*
	parse_free_chain() - free a linked list of parse_types
	This traverses a linked list and frees each element.
*/
void parse_free_chain(struct parse_type * temp_parse)
{
	struct parse_type * base_parse = temp_parse;
	void parse_free(struct parse_type * input_parse);
	
	while(temp_parse)
	{
		base_parse = temp_parse->next;
		temp_parse->next = NULL;
		parse_free(temp_parse);
		temp_parse = base_parse;
	}
}

/*
	parse_free() - frees parse_type data
	This simply frees the strings associated with
	a parse type, as well as the data itself.
*/
void parse_free(struct parse_type * input_parse)
{
	if(!input_parse)
		return;
	g_free(input_parse->input_string);
	input_parse->input_string = NULL;
	g_free(input_parse->output_string);
	input_parse->output_string = NULL;
	g_free(input_parse);
}

/*
	edit_delete_helper() - deletes a trigger or alias
	This function takes care of things common between deleting
	a trigger or alias, such as deleting the data from the
	linked list.
*/
void edit_delete_helper(struct conn_info * parent_conn, struct parse_type * parent_parse, struct parse_type * input_parse, int type)
{
	while(parent_parse)
	{
		if(parent_parse->next == input_parse)
			break;
		parent_parse = parent_parse->next;
	}
	
	if(!parent_parse) /* data to delete is first element */
	{
		switch(type)
		{
			case 1:
				parent_conn->aliases = input_parse->next;
				break;
			case 2:
				parent_conn->triggers = input_parse->next;
		}
	}
	else /* data to delete is not first element */
		parent_parse->next = input_parse->next;
}

/*
	edit_add_helper() - adds aliases or triggers to a linked list.
	This function takes care of those things common with adding
	data to a trigger or alias linked list.  It finds the last
	element in the list and appends new data to it, initializing
	it as well.
*/
void edit_add_helper(struct conn_info * input_conn, GtkWidget * clist, int type)
{
	struct parse_type * temp_parse = NULL;
	int find_last = TRUE;
		
	switch(type)
	{
		case 1:
			temp_parse = input_conn->aliases;
			if(!temp_parse) /* is there parse data already? */
			{ /* if not, create it */
				input_conn->aliases = g_malloc0(sizeof(parse_type));
				temp_parse = input_conn->aliases;
				find_last = FALSE;
			}
			break;
		case 2:
			temp_parse = input_conn->triggers;
			if(!temp_parse) /* is there parse_data already? */
			{ /* if not, create it */
				input_conn->triggers = g_malloc0(sizeof(parse_type));
				temp_parse = input_conn->triggers;				
				find_last = FALSE;
			}
			
	}

	if(find_last) /* do we need to find the last element? */
	{
		while(temp_parse->next)
			temp_parse = temp_parse->next;
		
		temp_parse->next = g_malloc0(sizeof(parse_type));
		/* ^^ make new data */
		temp_parse = temp_parse->next;
	}
	
	temp_parse->input_string = g_strdup("New parse rule"); /* initialize it */
	temp_parse->output_string = g_strdup("say I need to change this parse rule.\n");
	
	switch(type) /* reconstruct the clists */
	{
		case 1:
			init_parse_clist(input_conn->aliases, clist);
			break;
		case 2:
			init_parse_clist(input_conn->triggers, clist);
	}
	
	gtk_clist_select_row(GTK_CLIST(clist), gtk_clist_find_row_from_data(GTK_CLIST(clist), temp_parse), 0);
}

/*
	conn_copy() - copy connection data
	This is used to copy source to destination, make destination
	a unique copy of the given source.  Things like the
	state and socket are preserved.
*/
void conn_copy(struct conn_info * dest, struct conn_info * source)
{
	struct conn_info * next = dest->next;
	int descriptor = dest->descriptor, sock = dest->sock, state = dest->state;
	
	struct parse_type * parse_dup(struct parse_type * input_parse);
	void parse_free_chain(struct parse_type * temp_parse);
	void conn_free(struct conn_info * input_conn);
	struct log_info * log_dup(struct log_info * input_log);	
	void free_log_chain(struct log_info * input_log);
	
	if(!source) /* no source to copy */
		return;
	if(!dest) /* no place to copy */
		return;
		
	if(dest->host_name)
	{
		g_free(dest->host_name);
		dest->host_name = NULL;
	}
	dest->host_name = g_strdup(source->host_name);

	if(dest->mud_name)
	{
		g_free(dest->mud_name);
		dest->mud_name = NULL;
	}
	dest->mud_name = g_strdup(source->mud_name);

	dest->port = source->port;
	dest->next = next;

	if(dest->aliases)
	{
		parse_free_chain(dest->aliases);
		dest->aliases = NULL;
	}
	dest->aliases = parse_dup(source->aliases);

	if(dest->triggers)
	{
		parse_free_chain(dest->triggers);
		dest->triggers = NULL;
	}
	dest->triggers = parse_dup(source->triggers);

	if(dest->logs)
	{
		free_log_chain(dest->logs);
		dest->logs = NULL;
	}
	dest->logs = log_dup(source->logs);
	
	g_free(dest->font_name);
	dest->font_name = g_strdup(source->font_name);
	
	dest->font = source->font;
		
	dest->descriptor = descriptor;
	dest->sock = sock;
	dest->state = state;
}

/*
	general_entry_helper() - used to get text from an entry widget.
	This simply frees the char* passed and reinitialized it to
	the text in the entry box.
*/
void general_entry_helper(char * temp, GtkWidget * entry)
{
		g_free(temp);
		temp = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
}

/*
	process_user_input() - processes the user input (duh)
	This function finds out if the user typed an alias.
	If so, it processes that alias and continues.  If not,
	it sends the text to the mud.  This function is also
	called from within process_alias, making nested aliases
	possible.
*/
void process_user_input(char * text, struct parse_type * aliases, int socket_descriptor, GtkWidget * widget)
{
	char * temp_text, * temp_input;
	struct parse_type * parent = aliases;
	int i = 0;
	
	void process_alias(struct parse_type * alias, struct parse_type * parent, char * text, int socket_descriptor, GtkWidget * widget);
	void send_data(int socket_descriptor, char * text);
	void do_command(GtkWidget * widget, char * input);
	
	while(text[i] != ' ' && text[i] != '\0')
		i++;
	
	temp_text = g_malloc0(i + 1);
	strncpy(temp_text, text, i);

	if(*text == '#')
	{
		do_command(widget, text);
		return;
	}
	while(aliases) /* check all the aliases. */
	{
		temp_input = g_strdup(aliases->input_string);
		if(temp_input[strlen(temp_input) - 1] == '\n')
			temp_input[strlen(temp_input) - 1] = '\0';
		if(temp_text[0] == '\0' || temp_text[0] == '\n') /* end of list (?) */
			break;
		if(!(strcmp(temp_text, temp_input))) /* we've got a winner */
		{
			process_alias(aliases, parent, text, socket_descriptor, widget);
			return;
		}
		
		free(temp_input);
		aliases = aliases->next;
	}
	
	send_data(socket_descriptor, text); /* if no alias found, send it to the server */
	g_free(temp_text); /* garbage collection */
}

/*
	process_alias() - processes an alias
	This function finds all the '&' characters to see if they
	are arguments, and then applies the arguments to them.  It
	does check to make sure the argument exists, so that the user
	can't just type '&3434' for the heck of it.  It then sends the
	alias output text to process_user_input, which makes for the
	possibility of nested aliases.
*/
void process_alias(struct parse_type * alias, struct parse_type * parent, char * text, int socket_descriptor, GtkWidget * widget)
{
	struct argument_info * arguments, * base_args;
	char * temp_text, * new_text, * duplicate_text;
	int i = 0, j = 0;
	
	char * find_and_replace(struct argument_info * arguments, char * temp_text);
	
	arguments = g_malloc0(sizeof(argument_info));
	base_args = arguments;
	
	while(*text != ' ' && *text != '\0') /* skip whitespace */
		text++;
	
	if(*text != '\0' && *text != '\n') /* are we done? */
	{
		text++;
		new_text = text;
		arguments->argument = g_strdup(new_text); /* make argument '0' the entire string */
		
		while(*new_text != '\0')
		{
			while(*new_text != ' ' && *new_text != '\0')
			{
				new_text++;
				j++;
			}
			
			arguments->next = g_malloc0(sizeof(argument_info));
			arguments = arguments->next;
			
			arguments->argument = g_malloc0(j + 1);
			strncpy(arguments->argument, text, j);
			new_text++;
			text = new_text;
		}
	}
	
	duplicate_text = g_strdup(alias->output_string);
	
	/* find all the 'newlines', send each line individually
	   to process_user_input()
	*/
	while(1)
	{
		i = 0;
		temp_text = strsep(&duplicate_text, "\n");
		if(!temp_text)
			break;
			
		new_text = find_and_replace(base_args, temp_text); /* replace & with argument */
		process_user_input(new_text, parent, socket_descriptor, widget);
	}
	arguments = base_args;
	while(arguments) /* clean up */
	{
		if(arguments->argument)
			g_free(arguments->argument);
		base_args = arguments->next;
		g_free(arguments);
		arguments = base_args;
	}	
}

/*
	send_data() - send the text to the socket
	Simple enough.  Send the text.  If it didn't
	end in a newline, send a newline.
*/
void send_data(int socket_descriptor, char * text)
{
	send(socket_descriptor, text, strlen(text), 0);
	if(text[strlen(text) - 1] != '\n')
		send(socket_descriptor, "\n", 1, 0);
}

/*
	find_and_replace() - given a set of arguments and a string, find
	all the '&' and replace them with appropriate text.
	I really hate using a fixed size for the return_text, but I couldn't
	think of another way to do this.
*/
char * find_and_replace(struct argument_info * arguments, char * temp_text)
{
	int j = 0;
	char * return_text, * write_out;
	struct argument_info * base_args = arguments;
	
	return_text = g_malloc0(8192);

	write_out = return_text;
	while(*temp_text != '\0') /* end of string? */
	{
		if(*temp_text != '&') /* regular character? */
		{
			*write_out = *temp_text;
			temp_text++;
			write_out++;
			continue;
		}
		
		temp_text++;
		
		/* it was an ampersand.  Yipee! */
		/* is it a number or an asterisk? */
		if((!isdigit(*temp_text)) && (*temp_text != '*'))
		{
			*write_out = *temp_text;
			temp_text++;
			write_out++;
			continue;
		}
		
		if(*temp_text == '*') /* same as &0 */
		{				
			if(base_args)
			{
				strcat(return_text, base_args->argument);
				write_out = &return_text[strlen(return_text) - 1];
			}
			
			temp_text++;
			continue;
		}
		
		
		if(!arguments) /* no arguments. */
		{
			*write_out = *temp_text;
			temp_text++;
			write_out++;
			continue;
		}
		
		j = 0;
		
		while(arguments) /* while there are arguments */
		{
			if(arguments->argument)
			if(j == atoi(temp_text))
				break;
			arguments = arguments->next;
			j++;
		}
		
		if(arguments) /* did we find one? */
			if(arguments->argument)
			{
				strcat(return_text, arguments->argument);
				write_out = &return_text[strlen(return_text)];
			}
		while(isdigit(*temp_text)) /* skip past digit */
			temp_text++;
			
		arguments = base_args;
		continue;
	}
	return return_text;
}

/*
	pattern_match() - matches a pattern to a string
	This function uses regular expressions. I felt that
	this was a very good way to handle pattern matching.
	If the pattern matches the string, then the argument list
	is created.  A TRUE or FALSE is returned, depending on
	whether a match was found or not.
*/
int pattern_match(char * input, char * pattern, struct argument_info ** arguments)
{
	int status = FALSE, ret_value;
	regex_t re;
	regmatch_t matches[32];
	char * buf;
		
	struct argument_info * get_args(regmatch_t * matches, char * input, int num_matches);

	buf = g_strdup(pattern);
	if(buf[strlen(buf) - 1] == '\n') /* strip the newline */
		buf[strlen(buf) - 1] = '\0';

	ret_value = regcomp(&re, buf, REG_EXTENDED | REG_NEWLINE); /* compose the regex */
	if(buf)
		g_free(buf);
	
	if(!ret_value)
	{
		ret_value = regexec(&re, input, 32, matches, 0); /* do the comparison */
		if(ret_value != REG_NOMATCH)
		{
			*arguments = get_args(matches, input, 32); /* set the argument list */
			status = TRUE; /* found a match */
		}
	}
	return status;
}

/*
	get_args() - gets arguments from a set of regex matches.
	This function is here so that we can use find_and_replace with
	a trigger handler as well as the alias handler.  It simply goes
	through the matching strings and copies them to the argument list.
*/
struct argument_info * get_args(regmatch_t * matches, char * input, int num_matches)
{
	int i, len;
	struct argument_info * arguments = NULL, * base_args = NULL;
	
	for(i = 0; i < num_matches; i++)
	{
		len = matches[i].rm_so;
		if(len == -1)
			break;
		len = matches[i].rm_eo - len + 2;
		
		if(!arguments) /* no argument list yet, so create it */
		{
			arguments = g_malloc0(sizeof(argument_info));
			base_args = arguments; /* so we have something to return */
		}
		else /* we already have an argument list, so use it */
		{
			arguments->next = g_malloc0(sizeof(argument_info));
			arguments = arguments->next;
		}
		
		arguments->argument = g_malloc0(len);
		strncpy(arguments->argument, (char *)(input + matches[i].rm_so), len - 1);
		for(len = 0; len < strlen(arguments->argument); len++)
			if((arguments->argument[i] == '\n')|| (arguments->argument[i] == '\r'))
				arguments->argument[i] = '\0';
	}
	
	return base_args;
}

/*
	strip_escapes() - strips all escape sequences from the given text.
	This goes through a given string and strips all the escape and
	telnet sequences from it, so that the trigger handler can use it.
*/
void strip_escapes(char * input)
{
	char * write_out = input;
	char * read_in = input;
	
	while(*read_in != '\n' && *read_in != '\r' && *read_in != '\0')
		read_in++; /* skip newlines */
	
	while(*read_in != '\0') /* while we're in the string */
	{
		if(*read_in == '\xFF') /* telnet sequence */
		{
			read_in++; read_in++;
		}
		if(*read_in == '\x1B') /* ANSI escape sequence */
			while(*read_in != 'm' && *read_in != '\0')
				read_in++;
		else
			if(write_out != read_in)
			{
				*write_out = *read_in;
				write_out++;
			}
		
		read_in++;
	}
	
	write_out++;
	*write_out = '\0'; /* attach NULL to the end */
}

/*
	process_trigger() - find triggers given the input
	This function scans the input for triggers and acts
	on the triggers if they match.  This has been updated
	eliminate the possiblity of aliases in a trigger output
	string would not be executed.
*/
void process_trigger(struct conn_info * temp_conn, char * input, GtkWidget * widget)
{
	struct argument_info * arguments = NULL, * temp_args = NULL;
	char * temp_text = NULL, * duplicate_text, * new_text = NULL;
	struct parse_type * triggers = temp_conn->triggers;
	
	int pattern_match(char * input, char * pattern, struct argument_info ** arguments);
	void process_user_input(char * text, struct parse_type * aliases, int socket_descriptor, GtkWidget * widget);
	
	while(triggers) /* while we have triggers to match */
	{
		if(pattern_match(input, triggers->input_string, &arguments))
		{ /* ^^ do the matching */
			duplicate_text = g_strdup(triggers->output_string);
			/* send the output to the MUD, one line at a time */
			while(1)
			{
				temp_text = strsep(&duplicate_text, "\n");
				if(!temp_text)
					break;
			
				new_text = find_and_replace(arguments, temp_text); /* replace & with argument */
				process_user_input(new_text, temp_conn->aliases, temp_conn->sock, widget);
			}
			g_free(duplicate_text);
			
			g_free(temp_text);
			while(arguments) /* garbage collection */
			{
				if(arguments->argument)
						g_free(arguments->argument);

				temp_args = arguments;
				arguments = arguments->next;
				g_free(temp_args);
			}
			g_free(arguments);
		}
		triggers = triggers->next;
	}
}

/*
	do_command() - perform application commands from command input area.
	This looks to see if a valid command was entered in the input area,
	processing valid commands and flagging errant ones.
*/
void do_command(GtkWidget * window, char * command)
{
	struct conn_info * connection;
	
	void do_log(struct conn_info * connection, char * command);
	
	connection = gtk_object_get_data(GTK_OBJECT(window), "conn_info");
	
	if(!strncmp(command, "#reconnect", strlen("#reconnect")))
	{
		reconnect(window, NULL, NULL);
		return;
	}
	if(!strncmp(command, "#disconnect", strlen("#disconnect")))
	{
		disconnect_ask(window, NULL, NULL);
		return;
	}
	if(!strncmp(command, "#close", strlen("#close")))
	{
		close_conn(window, NULL, NULL);
		return;
	}
	if(!strncmp(command, "#help", strlen("#help")))
	{
		show_help(window, NULL, NULL);
		return;
	}
	if(!strncmp(command, "#edit", strlen("#edit")))
	{
		edit_selection_from_window(window, NULL, NULL);
		return;
	}
	if(!strncmp(command, "#log", strlen("#log")))
	{
		do_log(connection, command);
		return;
	}
	
	/* if we get here, none of the above commands were typed. */
	error_handle(17);
}

/*
	do_log() - maintains log windows and files.
	This slaps the text given in the #log command to the appropriate
	log file and/or window.
*/
void do_log(struct conn_info * connection, char * command)
{
	struct log_info * log;
	char * pos = command, * begin;
	FILE * logfile;
	GtkText * log_widget;
	int scroll = FALSE;
	GtkAdjustment * adj;
	
	if(!connection)
		return;
		
	while(*pos != ' ' && *pos != '\0' && *pos != '\n')
			pos++;
	pos++;
	begin = pos;
	
	pos++;
	while(*pos != ' ' && *pos != '\0' && *pos != '\n')
			pos++;
	log = connection->logs;
	
	while(log)
	{
		if(!strncmp(log->name, begin, pos - begin - 1))
		{
			pos++;
			if(log->log_file)
			{
				if(!(logfile = fopen(log->log_file, "a")))
				{
					error_handle(18);
					return;
				}
				fprintf(logfile, "%s\n", pos);
			}

			if(!log->log_widget && log->log_to_window)
				log->log_widget = create_log_window(connection->mud_name, log->name, log);

			if(log->log_to_window)
			{
				log_widget = gtk_object_get_data(GTK_OBJECT(log->log_widget), "text");
				adj = log_widget->vadj;
				if(adj->value == adj->upper - adj->lower - adj->page_size)
					scroll = TRUE;
				gtk_text_freeze(log_widget);
				gtk_text_insert(log_widget, NULL, NULL, NULL, pos, -1);
				gtk_text_insert(log_widget, NULL, NULL, NULL, "\n", -1);
				gtk_text_thaw(log_widget);
				if(scroll)
					gtk_adjustment_set_value(adj, adj->upper - adj->lower - adj->page_size);
			}

		break;
		}
		log = log->next;
	}
	
	if(log == NULL)
		error_handle(19);
}

/*
	init_log_clist() - initializes the log clist
	That about sums it up.  Initializes the clist for
	the edit window.
*/
void init_log_clist(struct log_info * log, GtkCList * clist)
{
	int i = 0;

	gtk_clist_clear(clist);
	
	if(!log)
		return;
		
	for(; log; log = log->next)
	{
		gtk_clist_append(clist, &log->name);
		gtk_clist_set_row_data(clist, i, log);
		i++;
	}
}

/*
	free_log() - frees a log structure
*/
void free_log(struct log_info * log)
{
	g_free(log->name);
	g_free(log->log_file);
}

/*
	free_log_chain() - frees a log chain
	This frees the current log, then traverses the
	log list until log == NULL
*/
void free_log_chain(struct log_info * log)
{
	struct log_info * temp_log;
	
	while(log)
	{
		temp_log = log->next;
		free_log(log);
		log = temp_log;
	}
}

/*
	init_mud_list() - initializes the mud clist for new muds
	This function will read in the file "~/.rmclist" and make a list
	out of the data.  If the file does not exist, it will be created, but
	left blank.
*/
void init_mud_list(GtkWidget * widget)
{
	FILE * infile;
	char * temp, * text, * list_text[3], * home;
	GtkWidget * clist;
	
	home = getenv("HOME");
	
	temp = g_malloc(strlen(home) + 10);
	
	strcpy(temp, home);
	strcat(temp, "/.rmclist");
	
	if(!(infile = fopen(temp, "r")))
	{
		if(!(infile = fopen(temp, "w+")))
		{
			g_free(temp);
			error_handle(20);
			return;
		}
	}
	g_free(temp);
	
	text = g_malloc0(512);

	clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "clist");
	fgets(text, 512, infile);
	temp = g_malloc(strlen("List created ") + strlen(text));
	strcpy(temp, "List created ");
	strcat(temp, text);
	
	gtk_window_set_title(GTK_WINDOW(widget), temp);
	
	while(!feof(infile))
	{
		fgets(text, 512, infile);
		home = g_strdup(text);
		list_text[0] = strsep(&home, "\t");
		list_text[0] = strsep(&home, "\t");
		list_text[1] = strsep(&home, "\t");
		list_text[2] = strsep(&home, "\t");
						
		gtk_clist_append(GTK_CLIST(clist), list_text);
	}
	
	g_free(text);
}

/*
	get_list_data() - socket input "callback" to handle reception of mudlist.
	This will get data from an already opened connection and dump it into
	the file ~/.rmclist for use in the MUD listing.  A status bar will also
	be updated so the user knows that something is going on.  This function
	exists to free the application from looking like it is "hanging" while
	downloading the list.
*/
void get_list_data(gpointer widget, int socket_descriptor, GdkInputCondition condition)
{
	int * tag, num_bytes, * size, fsize;
	struct stat filestat;
	char list_dump[8193], * home, * fname;
	FILE * outfile;
	GtkWidget * progress;
	float percent;
	
	void hang_up_list_conn(int tag, int descriptor);
	
	memset(list_dump, 0, 8193);
	
	tag = gtk_object_get_data(GTK_OBJECT(widget), "tag");
	size = gtk_object_get_data(GTK_OBJECT(widget), "size");
	progress = gtk_object_get_data(GTK_OBJECT(widget), "progress");
	
	num_bytes = recv(socket_descriptor, list_dump, 8192, 0); /* get the data */
	
	if(num_bytes < 1)
	{
		hang_up_list_conn(*tag, socket_descriptor);
		init_mud_list(gtk_object_get_data(GTK_OBJECT(widget), "parent"));
		delete_event(widget, 0, 0);
		g_free(tag);
		return;
	}

	home = getenv("HOME");
	
	fname = g_malloc(strlen(home) + 10);
	
	strcpy(fname, home);
	strcat(fname, "/.rmclist");
	
	fsize = stat(fname, &filestat);

	if(!fsize)
		fsize = (int)filestat.st_size;
	else
		fsize = 0;
		
	if(!(outfile = fopen(fname, "a")))
	{
		error_handle(22);
		hang_up_list_conn(*tag, socket_descriptor);
		delete_event(widget, 0, 0);
		g_free(tag);
		return;
	}
	
	percent = (float)fsize/(float)*size;
	
	gtk_progress_set_percentage(GTK_PROGRESS(progress), percent);
	fputs(list_dump, outfile);
	
	fclose(outfile);
	g_free(fname);
	g_free(home);
		
	return;
}

/*
	hang_up_list_conn() - hang up a connection associated with a mud list donwload.
	This just hangs up a connection and removes the listening tag.
*/
void hang_up_list_conn(int tag, int descriptor)
{
	close(descriptor);
	gdk_input_remove(tag);
	return;
}
